'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.mqttIpc = undefined;

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _mqtt = require('mqtt');

var _mqtt2 = _interopRequireDefault(_mqtt);

var _hash = require('../util/hash2.js');

var _emptyObj = require('../util/emptyObj.js');

var _prefix = require('../util/prefix.js');

var _shortid = require('shortid');

var _shortid2 = _interopRequireDefault(_shortid);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var mqttIpc = exports.mqttIpc = function mqttIpc(ipcSpec, lookup) {
  if (!ipcSpec.myId) {
    ipcSpec = _extends({}, ipcSpec, { myId: 'client/' + _shortid2.default.generate() });
  }
  var _ipcSpec = ipcSpec;
  var ipcUrl = _ipcSpec.ipcUrl;
  var myId = _ipcSpec.myId;


  var online = {}; // Which servers are active now?
  var mySubs = {}; // My subscriptions with sub messages.
  var peerSubs = {}; // Unsub handlers for peer subscriptions.
  var cmds = {}; // ipc commands
  var defCmd = function defCmd(cmd, handler) {
    return cmds[cmd] = handler;
  };

  var qidCounter = 0; // Message channel id:
  var nextQid = function nextQid() {
    return (++qidCounter).toString(36);
  };

  var dataTopic = 'data/' + myId;
  var okTopic = 'online/' + myId;
  var okAll = 'online/#';
  var qos = 1;var retain = true;

  var will = { topic: okTopic, payload: '', retain: retain, qos: qos };
  var client = _mqtt2.default.connect(ipcUrl, { will: will });

  var unsub = function unsub() {
    (0, _hash.forEach2)(mySubs, function (channel) {
      return stopChannel(channel);
    });
    (0, _hash.forEach2)(peerSubs, function (channel) {
      return stopChannel(channel);
    });
    client.publish(okTopic, '', { retain: retain, qos: qos }); // like will
    client.end();
  };

  client.on('connect', function () {
    client.subscribe(dataTopic, { qos: qos });
    client.subscribe(okAll, { qos: qos });
    client.publish(okTopic, 'ok', { retain: retain, qos: qos });
  });

  var send = function send(peerId, msg) {
    try {
      //// console.log(`${myId}==>${peerId}`,msg); ////
      msg = JSON.stringify(msg);
      client.publish('data/' + peerId, msg, { qos: qos });
    } catch (e) {
      console.log('failed send:', peerId, msg);
    }
  };

  var subscribe = function subscribe(peerId, qid, channel) {
    (0, _hash.wr2)(mySubs, peerId, qid, channel);
    if (online[peerId]) {
      startChannel(channel);
    }
  };
  var unsubscribe = function unsubscribe(peerId, qid, done, confirmed) {
    if (stopChannel((0, _hash.rd2)(mySubs, peerId, qid), done, confirmed)) {
      (0, _hash.del2)(mySubs, peerId, qid);
    }
  };

  var peerSubscribe = function peerSubscribe(peerId, qid, channel) {
    peerUnSubscribe(peerId, qid);
    (0, _hash.wr2)(peerSubs, peerId, qid, channel);
    startChannel(channel);
  };
  var peerUnSubscribe = function peerUnSubscribe(peerId, qid) {
    var channel = (0, _hash.del2)(peerSubs, peerId, qid);
    stopChannel(channel);
  };

  var connectFn = function connectFn(cSpec, recv) {
    var peerId = cSpec.peerId;
    var fname = cSpec.fname;
    var args = cSpec.args;

    var qid = nextQid();
    var ok = true;
    lookup(false, cSpec, function (_ref) {
      var yomo = _ref.yomo;
      var ctrl = _ref.ctrl;
      var fnDef = _ref.fnDef;

      var connect = function connect(args2) {
        return send(peerId, ['subscribe', myId, qid, fname, args2]);
      };
      var sendData = function sendData(data) {
        return send(peerId, ['cData', myId, qid, data]);
      };
      var info = {
        yomo: yomo, fname: fname, args: args, fnDef: fnDef, recv: recv,
        connect: connect, sendData: sendData,
        peerId: peerId, ipcSpec: ipcSpec, client: true
      };
      var channel = { info: info, ctrl: ctrl };
      ok && subscribe(peerId, qid, channel);
    });
    return function () {
      ok = false;
      unsubscribe(peerId, qid, true, false);
      send(peerId, ['unsubscribe', myId, qid]);
    };
  };

  var startChannel = function startChannel(channel) {
    if (1) {
      var _channel$info = channel.info;
      var peerId = _channel$info.peerId;
      var fname = _channel$info.fname;
      var args = _channel$info.args;
    }
    // On the client:
    // If the peer is now offline but gets online later,
    // startChannel will be called automatically.
    if (channel && online[channel.info.peerId]) {
      var start = channel.ctrl.start;
      var info = channel.info;
      var active = channel.active;

      if (active) {
        stopChannel(channel);
      }
      channel.active = true;
      channel.state = start(info);
      // ensure to call info.connect?(args) synchronously!
    }
  };

  var procChannel = function procChannel(channel, data) {
    if (channel) {
      var active = channel.active;
      var proc = channel.ctrl.proc;
      var info = channel.info;
      var state = channel.state;

      if (active) {
        channel.state = proc(info, data, state);
      }
    }
  };

  var stopChannel = function stopChannel(channel, done, confirmed) {
    if (channel) {
      var active = channel.active;
      var stop = channel.ctrl.stop;
      var info = channel.info;
      var state = channel.state;

      if (active) {
        stop(info, state);
      }
      channel.state = channel.active = undefined;

      if (done) {
        channel.done = true;
      }
      if (channel.done && (confirmed || !online[info.peerId])) {
        return true; // ok to delete the channel.
      }
      // a spurious confirmUnsub triggers a new startChannel:
      if (confirmed && !channel.done) {
        startChannel(channel);
      }
      return false; // don't delete the channel.
    }
  };

  defCmd('subscribe', function (_ref2) {
    var _ref3 = _slicedToArray(_ref2, 5);

    var __ = _ref3[0];
    var peerId = _ref3[1];
    var qid = _ref3[2];
    var fname = _ref3[3];
    var args = _ref3[4];

    lookup(true, { peerId: peerId, fname: fname, args: args }, function (_ref4) {
      var yomo = _ref4.yomo;
      var ctrl = _ref4.ctrl;
      var fnDef = _ref4.fnDef;

      var sendData = function sendData(data) {
        return send(peerId, ['sData', myId, qid, data]);
      };
      var info = {
        yomo: yomo, fname: fname, args: args, fnDef: fnDef, sendData: sendData,
        peerId: peerId, ipcSpec: ipcSpec, server: true
      };
      var channel = { info: info, ctrl: ctrl };
      peerSubscribe(peerId, qid, channel);
    });
  });
  defCmd('unsubscribe', function (_ref5) {
    var _ref6 = _slicedToArray(_ref5, 3);

    var __ = _ref6[0];
    var peerId = _ref6[1];
    var qid = _ref6[2];

    peerUnSubscribe(peerId, qid);
    send(peerId, ['confirmUnsub', myId, qid]);
  });
  defCmd('confirmUnsub', function (_ref7) {
    var _ref8 = _slicedToArray(_ref7, 3);

    var __ = _ref8[0];
    var peerId = _ref8[1];
    var qid = _ref8[2];

    unsubscribe(peerId, qid, false, true);
  });
  defCmd('sData', function (_ref9) {
    var _ref10 = _slicedToArray(_ref9, 4);

    var __ = _ref10[0];
    var peerId = _ref10[1];
    var qid = _ref10[2];
    var data = _ref10[3];

    var channel = (0, _hash.rd2)(mySubs, peerId, qid);
    if (!channel) {
      console.log('no channel1 for:', peerId, qid);
    }
    procChannel(channel, data);
  });
  defCmd('cData', function (_ref11) {
    var _ref12 = _slicedToArray(_ref11, 4);

    var __ = _ref12[0];
    var peerId = _ref12[1];
    var qid = _ref12[2];
    var data = _ref12[3];

    var channel = (0, _hash.rd2)(peerSubs, peerId, qid);
    if (!channel) {
      console.log('no channel2 for:', peerId, qid);
    }
    procChannel(channel, data);
  });

  client.on('message', function (topic, msg) {
    //// console.log(`@${myId} ${topic}: ${msg+''}`); ////
    if (topic === dataTopic) {
      try {
        msg = JSON.parse(msg.toString());
        var handler = cmds[msg[0]];
        if (handler) {
          handler(msg);
        } else {
          console.log('unknown cmd:', msg);
        }
        return;
      } catch (e) {
        return console.log('bad message', myId, msg, e);
      }
    }

    var peerId = (0, _prefix.afterPrefix)('online/', topic);
    if (peerId) {
      if (msg.toString() !== '') {
        // start-up: (re)subscribe
        if (!online[peerId]) {
          // un-bounce ok messages
          online[peerId] = true;
          var x = mySubs[peerId] || {};
          for (var qid in x) {
            startChannel(x[qid]);
          }
        }
      } else {
        // shut-down
        delete online[peerId];
        var _x = peerSubs[peerId];delete peerSubs[peerId];
        for (var _qid in _x) {
          stopChannel(_x[_qid]);
        }
        var y = mySubs[peerId];
        for (var _qid2 in y) {
          if (stopChannel(y[_qid2], false, false)) {
            delete y[_qid2];
          }
        }
        if ((0, _emptyObj.emptyObj)(y)) {
          delete mySubs[peerId];
        }
      }
    } else {
      console.log('unknown topic:', topic, msg + '');
    }
  });

  return { connectFn: connectFn, unsub: unsub };
};